#! /usr/local/bin/Rscript --vanilla

###############################################################################
# This is an example of how to use MetaDistance and some things you can do with
# the results.  You should probably be able to apply this to your own data by
# substituting your own input data, adjusting parameters, and removing parts
# of the analysis you don't need.

source("metadistance.R")

###############################################################################
# input files
matrix.file <- 'example_data.tsv'

# output files
plot.file <- 'metadist.pdf'
feature.weight.file <- 'feature-weight.tsv'
distance.matrix.file <- 'sample-sample-distance.tsv'

###############################################################################

pdf(plot.file)

###############################################################################
# Input data

d <- read.table(matrix.file, header=TRUE, sep='\t')

sample.id <- d[, 1]
class.label <- factor(d[, 2])
Y <- matrix(as.numeric(class.label))
X <- as.matrix(d[, 3:ncol(d)])

# Here are some transforms you might enjoy:
#X <- X / rowSums(X)
#X <- log(X + 0.001)
#X <- asin(sqrt(X))

###############################################################################
# Training

p <- 1 # should be set by cross-validation
result <- mdistrain(X, Y, 
  w0=matrix(1, ncol(X), 1),
  lam=200, # should be set by cross-validation
  k=3, # should be set by cross-validation
  p=p,
  tol=1e-5,
  maxiter=1000,
  verbose=3)

distances <- pdist(X, function(XI, XJ) my.weighted.pdist(result$w, XI, XJ, p))

###############################################################################
# Write feature weights to a file

write.table(result$w, sep='\t', quote=FALSE, col.names=FALSE, file=feature.weight.file)

###############################################################################
# Print non-zero weights to standard output in descending order

redor <- order(result$w, decreasing=TRUE)
print.me <- data.frame(feature=(1:nrow(result$w))[redor], weight=result$w[redor], name=(colnames(X)[redor]))
print.me <- print.me[print.me$weight != 0,]
write.table(print.me, row.names=FALSE, quote=FALSE, sep='\t')

###############################################################################
# Write distances to a file

distance.file <- file(distance.matrix.file, 'w')
for (i in 1:length(sample.id)) {
  for (j in 1:length(sample.id)) {
    cat(sprintf('%s\t%s\t%g\n', sample.id[i], sample.id[j], distances[i, j]), file=distance.file)
  }
}

###############################################################################
# CMDS visualization of distances

mds.fit <- cmdscale(as.dist(distances), eig=TRUE, k=20)

# plot relative sizes of eigenvalues, indicating portion of variance explained by each principal coordinate
barplot(mds.fit$eig / sum(mds.fit$eig), ylab='Portion of variance explained', xlab='Axis number')

class.colors <- rainbow(length(levels(class.label)))

# plot each of the top principal coordinates against the first
for (second.pc in 2:5) {
  plot(mds.fit$points[,1], mds.fit$points[,second.pc], col=class.colors[Y], xlab='Comp.1', ylab=sprintf('Comp.%d', second.pc))
  legend('bottomright', 0, levels(class.label), class.colors, title='Class')
}
