# This is the R implementation of MetaDistance.

# Authorship:  MetaDistance, its concepts and the original MATLAB
# implementation, are the work of Zhenqiu Liu.  This translation into R is
# primarily the work of Elliott Drabek, with participation by Brandi Cantarel.

# You will most likely use MetaDistance via the function mdistrain.  Depending
# on your data, you might also apply one or more of the transforms
# normalize.rows, arcsinVST, poissonVST, or negbinomialVST.

# Note to anyone modifying the code:  This code is meant to follow the Google R
# style guide.  Google it.

# Definitions:
#   An instance/feature matrix is a real-valued matrix with one row for each
#   instance and one column for each feature.

#########################################################################
# Miscellaneous helper functions

repmat <- function(a, n, m) {
  # replicates a matrix in repeated blocks
  #
  # Args:
  #   a:  the matrix to be replicated
  #   n:  the number of vertical copies
  #   m:  the number of horizontal copies
  #
  # Returns:  a matrix with n times as many rows and m times as many columns
  kronecker(matrix(1, n, m), a)
}

pround <- function(x, d) {
  # TODO(EFD):  describe this
  d <- round(d, digits=0)
  if (d < 1) {
    stop(sprintf('Number of digits must be positive integer d = %.4e\n', d))
  }
  if (x == 0 | is.nan(x) | is.infinite(x)) {
    x
  } else {
    p <- floor(log10(abs(x))) + 1
    factor <- 10 ^ (d - p)
    round(x * factor) / factor
  }
}

#########################################################################
# Transforms you might enjoy applying to the features before training
# Each of these takes a training feature matrix and returns a transformed
# version.  A training feature is a real-valued matrix with one row per
# instance and one column per feature.

normalize.rows <- function(X) {
  # normalize so that each row sums to 1
  # This is probably not what you want if your columns are not comprehensive.
  X / rowSums(X)
}

arcsinVST <- function(X) {
  asin(sqrt(X))
}

poissonVST <- function(X, methods=1) {
  if (methods == 1) {
    2 * sqrt(X + 3/8)
  } else {
    sqrt(X) + sqrt(X + 1)
  }
}

negbinomialVST <- function(X, L=10, methods=1) {
  if (methods == 1) {
    sqrt(L - 1/2) * asin(sqrt(X + 3/8) / sqrt(L - 3/4))
  } else {
    sqrt(L) * asin(sqrt(X/L)) + sqrt(L-1) * asin(sqrt(X + 3/4)/sqrt(L - 3/2))
  }
}

#########################################################################
# Core training routines

mdistrain <- function(x, y, w0=matrix(1, ncol(x), 1), lam=200, k=3, p=1, maxiter=1000, tol=1e-5, verbose=0) {
  # Trains a metadistance model from the raw training data, using tlsq as the
  # objective/gradient function.  This function is the main interface to
  # MetaDistance.  It is basically a wrapper around conjgradtrain.

  # Args:
  #   x:  The training features. It is a real-valued matrix, each column
  #     representing a feature, each row representing an instance.
  #   y:  The training target classes. It is an integer-valued matrix with one
  #     column and the same number of rows as x, each representing the class
  #     value of the instance in whose features are represented in the
  #     corresponding row of x.
  #   w0:  The starting value of the weight vector  It is a real-valued matrix
  #     with one row, and one column for each feature value.
  #   lam:  A penalty coefficient.  The higher its value, the stronger the
  #     pressure to produce a sparse weight vector.
  #   k:  The number of nearest neighbors considered in classification
  #   p:  The exponent to which the distance terms are raised. A value of 1
  #     gives Manhattan distance.  A value of 2 gives (squared) Euclidean
  #     distance.
  #   maxiter and tol:  Training ends after maxiter iteration or when the
  #     change in objective value or magnitude of the gradient is too small, as
  #     controlled by tol.
  #   verbose:  Higher values of verbose produce more diagnostic output.

  # Returns a list with the following components:
  #   w:  the final weight vector (of the same shape as w0)
  #   numiter:  the number of iterations actually done.

  Z <- MatrixInput(x, y, k, p)
  conjgradtrain(w0,
    function(...) mycgLineSearch(..., c2=1e-3),
    function(w) tlsq(w, Z$Zia, Z$Zir, lam),
    tol=tol, maxiter=maxiter, verbose=verbose)
}

MatrixInput <- function(Xtr, Ytr, k, p=1) {
  # Digests the raw training data into the form that will be used in training

  # Args:
  #   Xtr:  training feature matrix
  #   Ytr:  training class labels
  #   k:  number of nearest neighbors to consider (for intraclass distances)

  # Returns a list with the following components:
  #   Zia:  The intraclass matrix based on the k nearest neighbor
  #   Zir:  The interclass matrix with all points

  a <- sort(unique(Ytr))

  Zia <- c()
  Zir <- c()
  n <- length(a)
  for (y in 1:(n-1)) {
    X.in.class <- Xtr[Ytr == a[y],]
    X.out.of.class <- Xtr[Ytr > a[y],]
    Zia <- rbind(Zia, IntraKNN(X.in.class, k, p)$Yz)

    if (FALSE) {
      T1 <- repmat(X.in.class, nrow(X.out.of.class), 1)
      T2 <- repmat(X.out.of.class, nrow(X.in.class), 1) # this seems wrong
    } else {
      T1 <- repmat(X.in.class, nrow(X.out.of.class), 1)
      T2 <- c()
      for (i in 1:nrow(X.out.of.class)) {
        T2 <- rbind(T2, repmat(matrix(X.out.of.class[i,], nrow=1, ncol=ncol(X.out.of.class)), nrow(X.in.class), 1))
      }
    }
    Zir <- rbind(Zir, T1 - T2)
  }

  # Finally the class with largest label
  X.in.class <- Xtr[Ytr == a[n],]
  Zia <- rbind(Zia, IntraKNN(X.in.class, k, p)$Yz)

  Zia <- abs(Zia) ^ p
  Zir <- abs(Zir) ^ p

  list(Zia=Zia, Zir=Zir)
}

tlsq <- function(w, Zia, Zir, lam) {
  # Calculates the functional values and first order derivative for the
  # intra-product and inner-product distance matrix

  # Args:
  #   Zia:  the inter-class distance components
  #   Zir:  the intra-class distance components
  #   lam:  the penalty term

  # Returns a list with the following components:
  #   Fv:  the functional value for truncated least square
  #   F1:  the function for the first order derivative.

  L1 <- Zia %*% w - 1
  I1 <- ifelse(L1 > 0, 1, 0)
  E1 <- 0.5 * sum((L1 * I1) ^ 2)

  L2 <- 2 - Zir %*% w
  I2 <- ifelse(L2 > 0, 1, 0)
  E2 <- 0.5 * sum((L2 * I2) ^ 2)

  Fv <- E1 + E2 + lam * sum(w)

  X1 <- Zia * repmat(I1, 1, ncol(Zia))
  X2 <- Zir * repmat(I2, 1, ncol(Zir))

  F1 <- t(X1) %*% L1 - t(X2) %*% L2 + lam

  list(Fv=Fv, F1=F1)
}

pdist <- function(X, dist.function) {
  # calculates all-pairs distances

  # Args:
  #   X:  an instance-by-features matrix
  #   dist.function:  a function which computes distances between pairs of rows

  # Returns:
  #   A square matrix whose element i,j is the distance between row i and row j

  m <- nrow(X)
  result <- matrix(0, nrow=m, ncol=m)
  for (j in 1:(m - 1)) {
    for (i in (j + 1):m) {
      d <- dist.function(X[i,], X[j,])
      result[i, j] <- d
      result[j, i] <- d
    }
  }
  result
}

IntraKNN <- function(X, k, p=1) {
  # Finds the intraclass k nearest neighbor for each sample based on different
  # distance measure.  Neighbor can be repeated/non-repeated in this
  # implementation.

  # Args:
  #   X:  a sample/feature matrix for one class
  #   k:  the number of nearest neighbors to consider
  #   p:  The exponent to which the distance terms are raised. A value of 1
  #     gives Manhattan distance.  A value of 2 gives (squared) Euclidean
  #     distance.

  # Returns a list with the following components:
  #   Y:  the difference matrix
  #   Ind:  an array structure for neighbor structure
  #   Dist:  TODO(EFD)
  #   TODO(EFD) clarify the above

  z = pdist(X, function(XI, XJ) mypdist(XI, XJ, p))

  Ind = c()
  Yz = c()
  Dist = c()

  for (i in 1:nrow(X)) {
    tmp <- sort(z[i,], index.return=TRUE)
    nbdist <- tmp$x
    nbindex <- tmp$ix

    knb <- nbindex[2:(k+1)] # since the distance is zero for itself (k+1)
    kdist <- nbdist[2:(k+1)]

    ynb <- t(matrix(X[i,], ncol(X), k)) - X[knb,]

    Ind <- c(Ind, knb)
    Yz <- rbind(Yz, ynb)
    Dist <- c(Dist, kdist)
  }

  list(Yz=Yz, Ind=Ind, Dist=Dist)
}

mypdist <- function(XI, XJ, p) {
  # calculates the unweighted version of the distance between two feature rows

  # Args:
  #  XI, XJ:  the two feature rows
  #  p:  the exponent

  # Returns:

  sum(abs(XI - XJ) ^ p)
}

my.weighted.pdist <- function(w, XI, XJ, p) {
  # calculates the weighted version of the distance between two feature rows

  # Args:
  #  w:  the weight vector
  #  XI, XJ:  the two feature rows
  #  p:  the exponent

  # Returns:

  t(abs(XI - XJ) ^ p) %*% w
}

conjgradtrain <- function(x0, lsfun, ogfun, verbose=1, tol=1e-1, maxiter=1000, nu=0.1, abstol=0.00001, allowNonDecrease=0, digits=12) {
  # Searches for a vector to optimize the given objective function.
  # This is a revision (with non-negativity constraint) of  Polak-Ribiere
  # Nonlinear Conjugate Gradients for optimal solution, as described by "An
  # introduction to the conjugate gradient method without the agonizing pain"
  # by Shewchuk (1994) and "Numerical Optimization" by Nocedal and Wright.

  # Args:
  #   x0:  initial vector value to start search from
  #   lsfun:  line-search function
  #   ogfun:  objective/gradient function
  #   verbose:  Higher values of verbose produce more diagnostic output.
  #   tol:  geometric decrease in gradient magnitude to declare minimum
  #   maxiter:  stop after this many iterations (if no minimum found)
  #   nu: TODO(EFD)
  #   abstol:  stop if gradient magnitude goes below this
  #   allowNonDecrease:  don't stop if line search fails to find decrease
  #   digits:  digits of precision to use for objective comparisons

  # Returns a list with the following components:
  #   w:  the final weight vector (of the same shape as w0)
  #   numiter:  the number of iterations actually done.

  t0 <- proc.time()[3]
  ogcalls <- 0
  x <- x0
  x[x < 0] <- 0
  numiter <- 0
  j <- 0
  alpha <- 1e-10

  ogcalls <- ogcalls + 1
  ogfun.result <- ogfun(x)
  obj <- ogfun.result$Fv
  dx <- ogfun.result$F1
  r <- -dx
  s <- r
  d <- s
  deltanew <- t(r) %*% d
  deltazero <- deltanew

  min.deltanew = max(abstol, tol * abs(deltazero))

  repeat {
    if (numiter >= maxiter) {
      cat(sprintf('stopping after %d iterations because iteration threshold reached\n', numiter), file=stderr())
      break
    }
    if (abs(deltanew) <= min.deltanew) {
      cat(sprintf('stopping after %d iterations because magnitude of gradient is below threshold (%g <= %g)\n', numiter, sqrt(abs(deltanew)), sqrt(min.deltanew)), file=stderr())
      break
    }
    if (verbose >= 2) {
      cat(sprintf('iteration %d; objective = %g\n', numiter, obj), file=stderr())
    }

    numiter <- numiter + 1

    j <- j + 1
    prevobj <- obj
    if (alpha < 1e-10) {
      alpha <- 1e-10
    }
    lsfun.result <- lsfun(x, obj, dx, d, ogfun, alpha0=alpha, verbose=verbose)
    alpha <- lsfun.result$alpha
    obj <- lsfun.result$obj
    dx <- lsfun.result$dx
    ogc <- lsfun.result$ogc
    if (abs(obj - prevobj) < tol) {
      cat(sprintf('stopping after %d iterations because change in objective value is below threshold (%g < %g)\n', numiter, abs(obj - prevobj), tol), file=stderr())
      break
    }
    ogcalls <- ogcalls + ogc
    x <- x + alpha * d
    x[x < 0] <- 0
    r <- -dx
    deltaold <- deltanew
    deltamid <- t(r) %*% s
    deltanew <- t(r) %*% r
    beta <- (deltanew - deltamid) / deltaold
    d <- r + max(0, beta) * d
    if (deltamid / deltanew >= nu | t(d) %*% dx >= 0) {
      if (verbose >= 3 & length(x0) > 1) {
        cat('RESET\n', file=stderr())
      }
      d <- r
      j <- 0
    }
    s <- r
  }

  list(w=x, numiter=numiter)
}

update.state <- function(state, new.alpha) {
  # Sets the alpha value of a search state (as used in mycgLineSearch) and
  # updates other members of the structure according

  # Args:
  #  state:  The search state
  #  new.alpha:  The new value of alpha

  # Returns: the updated search state

  state$alpha <- new.alpha

  state$xx <- state$x0 + state$alpha * state$direction
  state$xx[state$zero.negatives & state$xx < 0] <- 0

  ogfun.result <- state$ogfun(state$xx)
  state$obj <- ogfun.result$Fv
  state$dx <- ogfun.result$F1
  state$ogcalls <- state$ogcalls + 1

  state
}

mycgLineSearch <- function(x0, obj0, dx0, direction, ogfun, seciter=5, alpha0=1e-8, c1=1e-4, c2=1e-1, digits=12, gamma=10, verbose=1) {

  # Quadratic interpolation line search with backtracking fallback.  Uses
  # Strong Wolfe termination conditions returns distance along direction
  # (alpha) and objective (obj) and gradient (dx) at x0+alpha.*direction

  # Args:
  # x0 - initial parameter value
  # obj0 - objective value at x0
  # dx0 - gradient at x0
  # direction - direction along which to find minimum
  # ogfun - pointer to objective/gradient calculation function

  #     x - parameter value at which to calculate obj/grad
  #     params - additional parameters
  #     obj - objective value at x
  #     dx - gradient at x
  # alpha - approximate minimum along direction is found at x0+alpha.*direction
  #
  # Additional parameters given by name/value pairs via varargin.
  # E.g. cgLineSearch(w, obj0, dx0, d, objfn, 'c2', 1e-2, 'verbose', 0)
  #
  # Secant Line Search algorithm taken from "An introduction to the
  # conjugate gradient method without the agonizing pain" by Jonathan
  # Shewchuk (1994).
  # Adapted by Zhenqiu Liu. UMB


  # x0:  location to start line search
  # direction:  direction from x0 along which to search
  # ogfun:  poitner to objective/gradient function
  # params:  cell array of parameters to pass to ogfun

  # alpha:  appx. location of minimum along direction
  #                 appx. minimium is: x0+alpha.*direction

  if (alpha0 <= 0) {
    stop('alpha0 must be greater than zero\n')
  }

  ##### Begin line search #####
  etazero <- as.vector(t(dx0) %*% direction)
  etaprev <- etazero

  state <- list(ogcalls=0, zero.negatives=TRUE, ogfun=ogfun, direction=direction, x0=x0)
  state <- update.state(state, alpha0)

  oldalpha <- state$alpha
  oldobj <- state$obj
  olddx <- state$dx

  # Make sure current step yields decrease in objective
  while ((pround(state$obj, digits) > pround(obj0 + c1 * state$alpha * etazero, digits)) && sum(x0 + state$alpha * direction != x0) > 0) {
    if (state$ogcalls > 1) {
      if (oldalpha > state$alpha / gamma && oldobj < obj0 + c1 * oldalpha * etazero) {
        state$alpha <- oldalpha
        state$obj <- oldobj
        state$dx <- olddx
        break
      }
    }
    state <- update.state(state, state$alpha / gamma)
  }

  beta <- state$alpha
  eta <- as.vector(t(state$dx) %*% direction)
  i <- 0
  # Try to find a point with lower directional derivative
  # Stop if too many iterations, large objective, no change, or small alpha
  while (abs(eta) > c2 %*% abs(etazero) && i < seciter && pround(state$obj, digits) <= pround(obj0, digits) && etaprev != eta && sum(state$xx != x0) > 0) {
    # Want newBeta/(eta-0) <- oldBeta/(etaprev-eta)
    beta <- eta * beta / (etaprev - eta)
    oldalpha <- state$alpha
    oldobj <- state$obj
    olddx <- state$dx

    new.alpha <- state$alpha + beta
    # Negative alpha could be a sign that initial parameter vector is bad
    if (new.alpha <= 0) {
      new.alpha <- 1
    } #
    etaprev <- eta
    i <- i + 1
    state <- update.state(state, new.alpha)
    eta <- as.vector(t(state$dx) %*% direction)
  } #

  # Find nonzeroalpha.
  if (sum(x0 + state$alpha * direction != x0) == 0) {
    new.alpha <- state $alpha
    while (x0 + new.alpha * direction == x0 && sum(abs(new.alpha * direction)) > 0) {
      new.alpha <- new.alpha * gamma
    }

    state <- update.state(state, new.alpha)
  }

  #backtrack
  while ((pround(state$obj, digits) > pround(obj0 + c1 * state$alpha * etazero, digits)) && sum(x0 + state$alpha * direction != x0) > 0) {
    if (state$ogcalls > 1) {
      if (oldalpha > state$alpha / gamma && oldobj < obj0 + c1 * oldalpha * etazero) {
        state$alpha <- oldalpha
        state$obj <- oldobj
        state$dx <- olddx

        break
      }
    }
    state <- update.state(state, state$alpha / gamma)
  }

  #checkConditions
  if (sum(state$xx != x0) == 0 & verbose) {
    cat('Warning:  Line search yielded no change in position.\nMay have reached limit of precision.\n')
  }

  state
}
